//+------------------------------------------------------------------+
//|                                                     moexcore.mqh |
//|                                    Copyright (c) 2025, Marketeer |
//|                           https://www.mql5.com/ru/articles/16964 |
//+------------------------------------------------------------------+
/*
   Main features:
   - Browse engines, markets, boards [,boardgroups]
   - List securities by market, by board, with filter by securitycollections
   - Global search securities by query, with filter by securitytypes, by securitygroups
   - Get security specs
   - Get security candles
   - Get security trades
   - Get security orderbook
   - Trading stats/aggregates
   
   How to use:
   - [Option for better results] Run 'moexopenapiedit.mq5' on OpenAPI json file of MOEX ISS (if available)
   - Run 'moexindexer.mq5' on API reference list, ISS index file (or get online from iss), and OpenAPI json
   - Use autogenerated moexref.mqh, moexindex.mqh, and /inc/ subfolder in your project, including dependencies

   NB: once a day download index and compare with local copy for important changes, for example, by jsoncmp.mq5
   NB: to swap output order use 'iss.reverse=true', or 'reversed=1', or superseded by 'sort_order', depending from request
*/

#include "toyjson2.mqh"
#include "moexlink.mqh"
#include "moexutils.mqh"
#include "StringUtils.mqh"
#include "htmlstrp.mqh"

//+------------------------------------------------------------------+
//| General enums                                                    |
//+------------------------------------------------------------------+

// NB: the term 'interval' is used in parameters and docs, but index/data defines it as 'duration'
enum MOEX_INTERVAL
{
  M1  =  1,
  M10 = 10,
  M60 = 60,
  H24 = 24,
  D7  =  7,
  D31 = 31,
  Q4  =  4
};

enum MOEX_FILTER_BY
{
   NONE,
   GROUP,
   TYPE
};

enum MOEX_SORT_ORDER
{
   asc,
   desc
};

enum MOEX_BOOL
{
   NO,
   YES
};

enum MOEX_YIELDDATETYPE
{
   MBS, MATDATE, OFFERDATE
};

enum MOEX_STATUS
{
   traded, nottraded, all
};

enum MOEX_TYPE
{
   daily, monthly
};

//+------------------------------------------------------------------+
//| General defines                                                  |
//+------------------------------------------------------------------+

#define MOEX_ALL (-1)  // unlimited download for all pages
#define MOEX_STRING(A) (#A + "=" + (string)(A))
#define MOEX_STRING_ENUM(A) (#A + "=" + EnumToMOEXString(A))
#define MOEX_STRING_TYPE(A,T) (#A + "=" + (string)(T)(A))
#define MOEX_STRING_BOOL(A) MOEX_STRING_TYPE(A,MOEX_BOOL)
#define MOEX_STRING_DATE(A) (#A + "=" + DateToMOEXString(A))
#define MOEX_STRING_TIME(A) (#A + "=" + TimeToMOEXString(A)) // TODO: apply to parameters with 'date-time' format (if any)

#define MOEX_IP(S,C,P) StringReplace(S,"[" + #P + "]",C(P))
#define MOEX_REPS(S,P) MOEX_IP(S, string, P)
#define MOEX_REPE(S,P) MOEX_IP(S, EnumToMOEXString, P)

//+------------------------------------------------------------------+
//| Main namespace                                                   |
//+------------------------------------------------------------------+

namespace Moex {

template<typename E1,typename E2>
E1 CheckRelation(E1 parent, E2 child, const int &ref[][2])
{
   E1 match = (E1)false;
   const int n = ArrayRange(ref, 0);
   if(!n) return parent;
   for(int i = 0; i < n; i++) // TODO: ArrayBSearch
   {
      if(ref[i][0] == child)
      {
         if(!parent || ref[i][1] == parent)
         {
            match = (E1)ref[i][1];
            if(!parent)
            {
               WARN_ONCE("%s '%s' is autodetected for %s '%s'", typename(E1), EnumToString(match), typename(E2), EnumToString(child));
            }
         }
         break;
      }
   }
   if(!match)
   {
      const string warning = StringFormat("%s '%s' does not belong to %s '%s'", typename(E2), EnumToString(child), typename(E1), EnumToString(parent));
      Alert(warning);
   }
   return match;
}

//+------------------------------------------------------------------+
//| Main struct for all autogenerated code                           |
//+------------------------------------------------------------------+

struct Base
{
protected:
   void checkRequirements(JsValue *obj, const string &required[])
   {
     if(!ArraySize(required)) return;
     
     for(int i = 0; i < ArraySize(required); i++)
     {
        if(obj[required[i]].t != JS_NULL) return; // at least one is present
     }
     
     Alert(StringFormat("Data for '%s' doesn't contain any of required sections: %s", _url, StringCombine(required, ',')));
   }

   // walk through the chain of blocks and assemble complete array of 'data'
   static JsValue *depaginate(AutoPtr<JsValue> &cache[], const int page1)
   {
      JsValue *all = cache[page1][];
      for(int i = page1 + 1; i < ArraySize(cache); i++)
      {
         JsValue *next = cache[i][];
         
         if(all.size() != next.size())
         {
            Alert("ERROR: Data structure mismatched in depagination:");
            string keys1[], keys2[];
            all.getProperties(keys1);
            next.getProperties(keys2);
            ArrayPrint(keys1);
            ArrayPrint(keys2);
            break;
         }
         
         for(int j = 0; j < fmin(all.size(), next.size()); j++) // interate through all sections
         {
            const int n = next[j]["data"].length();
            for(int k = 0; k < n; k++)
            {
               // FIXME: use ArrayInsert
               all[j]["data"].put(next[j]["data"][k]);
               next[j]["data"].drop(k);
            }
         }
      }
      
      ArrayResize(cache, page1 + 1); // pack the cache
      return all;
   }
   
   static string extractPlainText(const string html)
   {
      const int start = StringFind(html, "<body");
      const int stop = StringFind(html, "</body>");
      const int stoplen = StringLen("</body>");
      if(start > 0 && stop > 0)
      {
         HtmlStripper p;
         string text = StringSubstr(html, start, stop + stoplen - start);
         text = p.parse(text);
         StringTrimSpace(text);
         return text;
      }
      
      return NULL;
   }
   
   static void handleError(const string addr, const string text)
   {
      Alert("Problematic URL: ", addr);
      if(StringLen(text))
      {
         const string msg = extractPlainText(text);
         if(StringLen(msg)) Alert(msg);
      }
   }
   
public:

   string _url; // endpoint address and path with requested entities

   static JsValue *_fetch(string url)
   {
      static AutoPtr<JsValue> cache[]; // garbage collection and depagination
      
      if(StringLen(url) == 0)
      {
         ArrayResize(cache, 0);
         return (JsValue *)&JsValue::null;
      }
      
      const static string host = "https://iss.moex.com";
      const static string full = "start=-1";
      
      bool loop = StringReplace(url, full, "start=0") > 0;
      int start = 0;
      const int page1 = ArraySize(cache);
      
      do
      {
         const string text = link.fetch(host + url);
         const int code = link.getLastHttpCode();
      
         if(code == 200)
         {
            JsValue *obj = JsParser::jsonify(text);
            if(obj.t == JS_ERROR)
            {
               string buffer;
               obj.stringify(buffer);
               Alert(buffer);
               // try to find user-friendly message
               handleError(host + url, text);
            }
            int n = 0;
            // check for records in any of received data sections
            for(int i = 0; i < obj.size(); i++)
            {
               n = fmax(n, obj[i]["data"].length());
            }
            PUSH(cache, obj);
            if(loop) // if pagination was requested
            {
               loop = n; // continue while new data comes
               if(loop)
               {
                  const int got = start;
                  start += n;
                  loop = StringReplace(url, "start=" + (string)got, "start=" + (string)start) == 1;
               }
            }
         }
         else
         {
            handleError(host + url, text);
            if(StringLen(text))
            {
               const string errdump = "MOEX/" + TimeStamp() + ".err";
               WriteTextFile(errdump, text);
               Print("Received data saved: ", errdump);
            }
            break;
         }
      } while (loop && !IsStopped());
      
      if(ArraySize(cache) - page1 > 1)
      {
         return depaginate(cache, page1);
      }
      else if(ArraySize(cache) - page1 == 1)
      {
         return cache[page1][];
      }
      
      // return all we got so far or null
      return (JsValue *)&JsValue::null;
   }
   
   static MoexLink link;
};

//+------------------------------------------------------------------+
//| Default struct for no-name entities                              |
//+------------------------------------------------------------------+

struct Entity: public Base
{
   JsValue *_get(const string query = NULL)
   {
      return _fetch(_url  + ".json" + link.buildQuery(query));
   }
};

static MoexLink Base::link;

} // namespace

//+------------------------------------------------------------------+
